
#include "Ai.h"
#include "Game.h"
#include "StringData.h"
#include "StraightData.h"
#include "VirtualStraight.h"
#include "CheckReach.h"

#include <set>
#include <algorithm>

#include <windows.h>
#include <assert.h>


#ifndef _DEBUG
# define IrenseiPrintf
#endif


namespace Irensei {


//																		//
//								`h								//
class AI {
public:
	AI( const IRENSEI_RULE_DESCRIPTION & rule, Game::EBlackOrWhite side );

	int Think( IRENSEI_BOARD_POSITION & result, const IRENSEI_BOARD_POSITION * lastMove, HANDLE hout );

	int Undo( IRENSEI_BOARD_POSITION & lastMove, HANDLE hout );

	int Debug( IRENSEI_DEBUG_BOARD & dest, int idx, HANDLE hout );
	static int CountDebug( void )		{ return 6; };

	struct Position {
		Position( int a, int b ) : x(a), y(b)	{}
		Position( const StraightData::Position & o ) : x(o.x), y(o.y)	{}
		int x;
		int y;
		bool operator < ( const Position & o ) const { return x<o.x || (x==o.x && y < o.y); }
		operator const IRENSEI_BOARD_POSITION & ( void ) const { return *reinterpret_cast<const IRENSEI_BOARD_POSITION*>(this); }
		operator IRENSEI_BOARD_POSITION & ( void )  { return *reinterpret_cast<IRENSEI_BOARD_POSITION*>(this); }
		bool isOuterOf( int outer/* line count*/ ) const {
			return x < outer || x >= Board::MaxX - outer || y < outer || y >= Board::MaxY - outer;
		}
	};

private:
	AI();		// ftHgRXgN^𖳌

	Game game;
	Game::EBlackOrWhite myside;

	const IRENSEI_RULE_DESCRIPTION rule;

	IRENSEI_DEBUG_BOARD legalbuf;
	IRENSEI_DEBUG_BOARD straightLength;
	IRENSEI_DEBUG_BOARD scoreForDebug;
	IRENSEI_DEBUG_BOARD myReach;
	IRENSEI_DEBUG_BOARD hisReach;
	IRENSEI_DEBUG_BOARD VStrLength;


	// vlnA[eBeB[
	void makeLegalMoves( std::set< Position > & legalMoves, IRENSEI_DEBUG_BOARD  & legalbuf ) const;
	void makeStraightLength( IRENSEI_DEBUG_BOARD & straightLength, const StraightData & strings ) const;
	void makeVirtualStraightLength( IRENSEI_DEBUG_BOARD & virtualLength, const VirtualStraightData & vstrs, Game::EBlackOrWhite side ) const;
	int randomMove( IRENSEI_BOARD_POSITION & result, const IRENSEI_DEBUG_BOARD & legalbuf, HANDLE hout ) const;
	int searchWin( IRENSEI_BOARD_POSITION & result, HANDLE hout ) const;
	int stop6( IRENSEI_BOARD_POSITION & result, HANDLE hout ) const;
	int stop5( IRENSEI_BOARD_POSITION & result, HANDLE hout ) const;
	int make6( IRENSEI_BOARD_POSITION & result, const StraightData & straights, HANDLE hout ) const;
	int beginning( IRENSEI_BOARD_POSITION & result, HANDLE hout ) const;		// 
	int evaluate( IRENSEI_BOARD_POSITION & result, HANDLE hout )/* not const */;
	void evalvstr( int scores[ Board::MaxY ][ Board::MaxX ], HANDLE hout ) const;
	static Board::EStone get( const Board * board, const Position & pos )	{ return board->Get( pos.x, pos.y ); }
	static Board::EStone get( const Board * board, int x, int y )	{ return board->Get( x, y ); }
	static void SetResult( IRENSEI_BOARD_POSITION & result, const Position & pos )	{ result.x = pos.x; result.y = pos.y; }
	static void SetResult( IRENSEI_BOARD_POSITION & result, int x, int y )	{ result.x = x; result.y = y; }
};


AI::AI( const IRENSEI_RULE_DESCRIPTION & rl, Game::EBlackOrWhite sd )
		:
	myside(sd),
	rule(rl)
{
	game.SetRule( rule );

	srand(GetTickCount());
}

int AI::Undo( IRENSEI_BOARD_POSITION & lastMove, HANDLE hout )
{
	assert( 0 != game.CountMove() );
	game.Undo();

	const Game::Move lm = game.GetLastMove();
	lastMove.x = lm.x;
	lastMove.y = lm.y;

	return 1;
}

int AI::Debug( IRENSEI_DEBUG_BOARD & dest, int idx, HANDLE hout )
{
	switch ( idx ) {
		case 0:
			memcpy( &dest, &legalbuf, sizeof(dest) );
			break;
		case 1:
			memcpy( &dest, &straightLength, sizeof(dest) );
			break;
		case 2:
			memcpy( &dest, &scoreForDebug, sizeof(dest) );
			break;
		case 3:
			memcpy( &dest, &myReach, sizeof(dest) );
			break;
		case 4:
			memcpy( &dest, &hisReach, sizeof(dest) );
			break;
		case 5:
			memcpy( &dest, &VStrLength, sizeof(dest) );
			break;
		default:
			return 0;
	}

	return 1;
}


//																		//
//								vl[`							//

static int count_stones( const Board * board, int stoneMask )
{
	int sum = 0;
	for ( int y = 0; y < Board::MaxY; y++ ) {
		for ( int x = 0; x < Board::MaxX; x++ ) {
			if ( board->Get(x, y) & stoneMask )
				sum++;
		}
	}
	return sum;
}

int AI::Think( IRENSEI_BOARD_POSITION & result, const IRENSEI_BOARD_POSITION * lastMove, HANDLE hout )
{
	if ( lastMove )
		game.NewMove( lastMove->x, lastMove->y );
	IrenseiPrintf( hout, "%d th move: ", game.CountMove() + 1 );

	result.x = -1;

	// @̗
	std::set< Position > legalMoves;
	makeLegalMoves( legalMoves, legalbuf );

	// A쐬
	StringData strings( game.GetBoard() );

	// я쐬
	StraightData straights( game.GetBoard() );
	makeStraightLength( straightLength, straights );

	// VA\e[u쐬
	CheckReach( myReach, game.GetBoard(), rule, myside );
	CheckReach( hisReach, game.GetBoard(), rule, game.NextOrder(myside) );

	VirtualStraightData vstrs( game.GetBoard(), rule.lineLength + 1/*ԒAp*/ );
	makeVirtualStraightLength( VStrLength, vstrs, myside );

	memset( scoreForDebug.stones, -128, sizeof(scoreForDebug.stones) );

	// vl
	do/* break  */ {
		if ( searchWin( result, hout ) )		// ɂȂ_Αł
			break;

		if ( stop6( result, hout ) )			// U~߂
			break;

		if ( make6( result, straights, hout ) )	// BUȂł
			break;

		if ( stop5( result, hout ) )			// BT~߂
			break;

		if ( beginning( result, hout ) )		// 
			break;

		if ( evaluate( result, hout ) )			// ]֐
			break;

		// _ɑłiłĂꏊ̂ɂɓ̓C}C`Ȃ߁AƂ肠j
		if ( randomMove( result, legalbuf, hout ) ) {
			IrenseiPrintf( hout, "moved in random" );
			break;
		}
	} while ( 0 )/* break  */;

	if ( result.x != -1 ) {
		game.NewMove( result.x, result.y );
		return 1;
	}

	// 
	return 0;
}


// ɂȂ_Αł
int AI::searchWin( IRENSEI_BOARD_POSITION & result, HANDLE hout ) const
{
	for ( int y = 0; 19 > y; y++ ) {
		for ( int x = 0; 19 > x; x++ ) {
			if ( legalbuf.stones[y][x] == Game::EMResultWin ) {
				result = Position( x, y );
				IrenseiPrintf( hout, "I win!" );
				return 1;
			}
		}
	}
	return 0;
}


// U~߂
int AI::stop6( IRENSEI_BOARD_POSITION & result, HANDLE hout ) const
{
	int outer = rule.outerInvalidLines;
	for ( int y = outer; y < Board::MaxY - outer; y++ ) {
		for ( int x = outer; x < Board::MaxX - outer; x++ ) {
			if ( hisReach.stones[y][x] == 0 ) {
				if ( legalbuf.stones[y][x] <= Game::EMResultWin ) {		// @H
					result.x = x;
					result.y = y;
					IrenseiPrintf( hout, "should stop the 6." );
					return 1;
				}
				else {
					IrenseiPrintf( hout, "I'll lose..." );
				}
			}
		}
	}
	return 0;
}
// BT~߂
int AI::stop5( IRENSEI_BOARD_POSITION & result, HANDLE hout ) const
{
	const int outer = rule.outerInvalidLines;
	const int mx = Board::MaxX;
	const int my = Board::MaxY;
	std::vector<Position> movesShouldDo;
	for ( int y = outer; y < Board::MaxY - outer; y++ ) {
		for ( int x = outer; x < Board::MaxX - outer; x++ ) {
			if ( hisReach.stones[y][x] == 1 ) {
				if ( legalbuf.stones[y][x] <= Game::EMResultWin ) {		// @H
					// {Ɏ~܂H
					if (	(y>0  && x>0  && straightLength.stones[y-1][x-1] == 5 ) ||
							(y>0  && x>=0 && straightLength.stones[y-1][x  ] == 5 ) ||
							(y>0  && x<mx && straightLength.stones[y-1][x+1] == 5 ) ||
							(y>=0 && x>0  && straightLength.stones[y  ][x-1] == 5 ) ||
							(y>=0 && x>=0 && straightLength.stones[y  ][x  ] == 5 ) ||
							(y>=0 && x<mx && straightLength.stones[y  ][x+1] == 5 ) ||
							(y<my && x>0  && straightLength.stones[y+1][x-1] == 5 ) ||
							(y<my && x>=0 && straightLength.stones[y+1][x  ] == 5 ) ||
							(y<my && x<mx && straightLength.stones[y+1][x+1] == 5 )
					) {
						// ǉ
						movesShouldDo.push_back( Position( x, y ) );
					}
				}
			}
		}
	}
	if ( movesShouldDo.empty() )
		return 0;
	int r = rand() * movesShouldDo.size() / (RAND_MAX+1);
	result.x = movesShouldDo[r].x;
	result.y = movesShouldDo[r].y;
	IrenseiPrintf( hout, "should stop the 5." );
	return 1;
}


// UȂł
int AI::make6( IRENSEI_BOARD_POSITION & result, const StraightData & _straights, HANDLE hout ) const
// straight ݂̂̃o[WB
// virtual straight ɂ̂ɂ邱Ƃ]܂
// ɕGɂȂĂ邪AvɁA
//	PjTAUAɂȂ邩A
//	QjUɂȂƂAǂ󔒂iЕ~ł́AɎ~߂j
//	Rjǂ̋󔒂ւł͂ȂiԂ̒AɂȂȂj
//	SjǂłVɂĂAL͈͓
// ƂƂŁÂ\GB
{
	const Board * board = game.GetBoard();
	const int outer = rule.outerInvalidLines;
	const Board::EStone me = Game::OrderToStone( myside );
//	const Board::EStone other = ~me;

	const std::set< StraightData::Straight > & straights = _straights.straights;
	std::set< StraightData::Straight >::const_iterator itor;
	for ( itor = straights.begin(); itor != straights.end() && itor->length >= 5; itor++ ) {
		if ( itor->length > 5 )
			continue;		// TÂ݂ΏۂƂ
		if ( itor->color != me )
			continue;		// ̂ˁ[I
		StraightData::delta d = StraightData::GetDelta( itor->dir );
		StraightData::Position p;
		// т̂ǂ炩̐ɐ΂ 6 ɂȂȂ
		if ( itor->firststop != Board::EStoneSpace || itor->laststop != Board::EStoneSpace )
			continue;
		// stop ̈ʒu͈͂Ȃʖ
		if ( (itor->lastend + d).isOuterOf( outer ) )
			continue;
		if ( (itor->lastend - d).isOuterOf( outer ) )
			continue;
		// ԂȂAт̂ǂ炩̐Q󔒂̐΂Ŗ΂ȂȂ
		if ( myside != Game::EBWBlack ) {
			const StraightData::Position * end = &itor->lastend;
			for ( int lpc = 0; 2 > lpc; lpc++ ) {
				p = *end + d + d;
				if ( ! p.isOuterOf( outer ) ) {
					result.x = end->x + d.x;
					result.y = end->y + d.y;
					if ( game.CheckLegalMove( game.GetBoard(), result.x, result.y, myside ) <= Game::EMResultWin ) {
						IrenseiPrintf( hout, "I'll win(white)." );
						return 1;
					}
				}
				d = d * -1;
				end = &itor->firstend;
			}
			continue;
		}
		// ԂȂA
		else {
			const StraightData::Position * thisend = &itor->lastend;
			const StraightData::Position * anotherend = &itor->firstend;
			for ( int lpc = 0; 2 > lpc; lpc++ ) {
				p = *thisend + d + d;		// Q
				if ( ! p.isOuterOf( outer ) ) {
					const Board::EStone nextstop = board->Get( p.x, p.y );
#if 0/*ݍӂƂ*/
					if ( nextstop & other )
						;		// ̐΂΂UɂȂȂ
					else if ( nextstop == me )
						;		// Ɏ̐΂ȂAAɂȂiȂȂȂ searchWin() ̊Ǌj
#else/*悤ɂ*/
					if ( nextstop != Board::EStoneSpace )
						;		// UɂȂȂAɂȂ
#endif
					else {
						p = *anotherend - d - d;	// Q낪
						if ( board->Get( p.x, p.y ) == me )
							;		// ̐΂ȂAAɂȂiȂȂȂ searchWin() ̊Ǌj
						else {
							// BUI
							result.x = thisend->x + d.x;
							result.y = thisend->y + d.y;
							if ( game.CheckLegalMove( game.GetBoard(), result.x, result.y, myside ) <= Game::EMResultWin ) {
								IrenseiPrintf( hout, "I'll win(black)." );
								return 1;
							}
						}
					}
				}
				d = d * -1;
				thisend = &itor->firstend;
				anotherend = &itor->lastend;
			}
		}
	}
	return 0;
}


// 
int AI::beginning( IRENSEI_BOARD_POSITION & result, HANDLE hout ) const
{
	if ( game.CountMove() > 5 )
		return 0;		// ՂȂ

	if ( game.CountMove() == 0 ) {
		// ́AV܂ށAE 4*4 ͈̔
		result.x = Board::MaxX/2 + rand() * 4 / RAND_MAX;
		result.y = Board::MaxY/2 - rand() * 4 / RAND_MAX;
		IrenseiPrintf( hout, "that's beginning." );
		return 1;
	}

	if ( game.CountMove() == 1 ) {
		// ԍŏ̈
		const Game::Move firstMove( game.GetLastMove() );		// ԏ
		if ( firstMove.x == Board::MaxX/2 && firstMove.y == Board::MaxY/2 ) {
			result.x = firstMove.x + (rand() * 3 / (RAND_MAX+1) - 1)/*[-1..1]*/;
			result.y = firstMove.y + (rand() * 3 / (RAND_MAX+1) - 1)/*[-1..1]*/;
			if ( result.x == Board::MaxX/2 && result.y == Board::MaxY/2 ) {
				result.x = firstMove.x + 1;
				result.y = firstMove.y;
			}
		}
		else {
			struct near_center {
				int operator () ( int center, int v ) {
					if      ( v == center )	return (rand() * 3 / (RAND_MAX+1) - 1)/*[-1..1]*/;
					else if ( v < center )	return 1;
					else 					return -1;
				}
			};
			result.x = firstMove.x + near_center()( Board::MaxX/2, firstMove.x );
			result.y = firstMove.y + near_center()( Board::MaxY/2, firstMove.y );
		}
		return 1;
	}

	return 0;
}



// ]֐
int AI::evaluate( IRENSEI_BOARD_POSITION & result, HANDLE hout )/* not const */
{
	StringData orgstrings( game.GetBoard() );

	const Board::EStone me = game.OrderToStone( game.NextOrder() );
	const int other = me ^ -1;

	int scores[ Board::MaxY ][ Board::MaxX ];
	int intmin = 0x80000000;	// Q̕␔\łRQrbg̏ꍇ̍ŏl
	for ( int lpc = 0; lpc < Board::MaxX*Board::MaxY; lpc++ )
		scores[0][lpc] = intmin;
	evalvstr( scores, hout );

	int max = intmin;
	int cntmax = 0;
	for ( int y = 0; y < Board::MaxY; y++ ) {
		for ( int x = 0; x < Board::MaxX; x++ ) {
			if ( legalbuf.stones[y][x] > Game::EMResultWin ) {
				scores[y][x] = intmin;		// evalstr() Ȃ
				continue;
			}

			// ꂪȂAՖʂlĂ݂
			const int lastOtherStones = count_stones( game.GetBoard(), other );
			game.NewMove( x, y );		// @@ 
			const_cast<Board*>(game.GetBoard())->ChangeStone( x, y, Board::EStoneSpace );
			const Board * board = game.GetBoard();
			const int curOtherStones = count_stones( game.GetBoard(), other );
			const bool killedAny = curOtherStones < lastOtherStones;

			StringData tmpstrings;
			IRENSEI_DEBUG_BOARD tmpmyreach;
			IRENSEI_DEBUG_BOARD tmphisreach;
			if ( killedAny ) {
				tmpstrings.Remake( game.GetBoard() );
				CheckReach( tmpmyreach, game.GetBoard(), rule, myside );
				CheckReach( tmphisreach, game.GetBoard(), rule, game.NextOrder(myside) );
			}
			const StringData & strings = killedAny ? tmpstrings : orgstrings;
			const IRENSEI_DEBUG_BOARD & myreach = killedAny ? tmpmyreach : myReach;
			const IRENSEI_DEBUG_BOARD & hisreach = killedAny ? tmphisreach : hisReach;

			// ̘A\
			int mr = myreach.stones[y][x];
			mr = ( mr == -128 ) ? rule.lineLength : mr;
			mr = ( mr < 0 ) ? (rule.lineLength+mr)*1 : ((rule.lineLength-mr)*4);
			// ̘A\
			int hr = hisreach.stones[y][x];
			hr = ( hr == -128 ) ? rule.lineLength : hr;
			hr = ( hr < 0 ) ? (rule.lineLength+hr)*2 : ((rule.lineLength-hr)*3);
			// ̐ΐi̕j̓_BAɂȂƁj
			int mt = 0;		// c
			int mx = 0;		// ΂
			int ht = 0;
			int hx = 0;
			const Board::EStone out = Board::EStoneOutOfBounds;
			Board::EStone stone;
			stone = board->Get(x-1,y-1);	if (stone!=out) { mx += (stone == me);	hx += (stone & other); }
			stone = board->Get(x  ,y-1);	if (stone!=out) { mt += (stone == me);	ht += (stone & other); }
			stone = board->Get(x+1,y-1);	if (stone!=out) { mx += (stone == me);	hx += (stone & other); }
			stone = board->Get(x-1,y  );	if (stone!=out) { mt += (stone == me);	ht += (stone & other); }
			stone = board->Get(x+1,y  );	if (stone!=out) { mt += (stone == me);	ht += (stone & other); }
			stone = board->Get(x-1,y+1);	if (stone!=out) { mx += (stone == me);	hx += (stone & other); }
			stone = board->Get(x  ,y+1);	if (stone!=out) { mt += (stone == me);	ht += (stone & other); }
			stone = board->Get(x+1,y+1);	if (stone!=out) { mx += (stone == me);	hx += (stone & other); }
			int ms = 0;
			int hs = 0;
			if ( mt == 0 && ht == 3 )
				ms = -8;				// 荞
#if 0
			if      ( mc == 1 ) {
				if ( hc == 0 )			// LтȂ
					ms = 4;
			}
			else if ( mc == 2 ) {	// тA΂
				if ( hc != 0 )
					ms = 5;				// `Ă
			}
			else if ( mc == 3 ) {
				if ( hc == 0 )
					ms = -5;			// `ĂȂJPcM`
				else
					ms = 2;				// `Ă
			}
			if      ( hc == 1 ) {	// t
				if ( mc == 0 )
					hs = 4;
			}
			else if ( hc == 2 ) {
				if ( mc != 0 )
					hs = 3;
			}
			else if ( hc == 3 ) {
				if ( mc == 1 )
					hs = 1;				// `Ɍp
				else
					hs = -5;			// zER~H
			}
#endif
			// Ԃ͍]
			int es = 0;
			StringData::Position here( x, y );
			for ( std::vector<StringData::String>::const_iterator itor = strings.strings.begin(); itor != strings.strings.end(); itor++ ) {
				if ( itor->color != me )
					continue;
				if ( itor->breathPoints.size() == 1 && here == *itor->breathPoints.begin() ) {
					if ( strings.CheckSityou( x, y, me ) ) {
						es = -100;		// V`E̓o͍Œ
						IrenseiPrintf( hout, "(%d %d) ̓V`E̓o\r\n", x, y );
						goto endBreath;
					}
					else {
						// PqĂꂽƂ肠
						IrenseiPrintf( hout, "(%d %d) Pq𓦂o\r\n", x, y );
						es = 32;
						goto endBreath;
					}
				}
/*
				else if ( _H )
					es = 10;
*/
				else {
					const int breathMagic = 2/*@@MAGIC!!*/;
					if (   (itor->breathPoints.size() <= 3/*@@MAGIC!!*/) ||
						(itor->breathPoints.size() < (itor->stones.size() / 2 - breathMagic))
					) {	// アAH
						for ( std::set<StringData::Position>::const_iterator pos = itor->breathPoints.begin(); pos != itor->breathPoints.end(); pos++ ) {
							if ( *pos == here ) {
								const int s = hs + 5 * (breathMagic + itor->stones.size() - itor->breathPoints.size());
								es += s;
								IrenseiPrintf( hout, "アA̓o (%d %d - %d) ", x, y, s );
							}
						}
					}
				}
			}
endBreath:

			// 
			int gs = lastOtherStones - curOtherStones;
			if ( killedAny ) {
				IrenseiPrintf( hout, "(%d %d) %dq ", x, y, gs );
				gs = gs*gs / 2 + 4;		// q 4, 2..6, 3..8, 4..12
				// @@ ł邩ǂB
			}

			// `

			int score = mr + hr + ms + hs + es + gs + (( scores[y][x] == intmin ) ? 0 : scores[y][x]);
			if ( score <= 0 )
				score = 0;
			if ( max < score ) {
				max = score;
				cntmax = 1;
			}
			else if ( max == score ) {
				cntmax++;
			}
			scores[y][x] = score;

			game.Undo();		// @@  NewMove() 
		}
	}
	IrenseiPrintf( hout, "]l max = %d, cntmax = %d ", max, cntmax );
	if ( max == intmin )
		return 0;

	std::vector< Position > testMoves;
	std::vector< int > cofs;
	testMoves.reserve( 19*19 );
	cofs.reserve( 19*19 );
	int total = 0;
	for ( int y = 0; y < Board::MaxY; y++ ) {
		for ( int x = 0; x < Board::MaxX; x++ ) {
			int score = scores[y][x];
			scoreForDebug.stones[y][x] = (score <= 0) ? -128 : score * 99 / max;
			if ( score == max ) {
				total += score;
				cofs.push_back( total );
				testMoves.push_back( Position(x,y) );
			}
		}
	}
	if ( testMoves.size() ) {
		int r = total * rand() / (RAND_MAX + 1);
		std::vector< int >::const_iterator i = std::lower_bound( cofs.begin(), cofs.end(), r );
		*reinterpret_cast<Position*>(&result) = testMoves[ i - cofs.begin() ];
		return 1;
	}
	return 0;
}


void AI::evalvstr( int scores[ Board::MaxY ][ Board::MaxX ], HANDLE hout ) const
// sȈʒuɕ]lL^Ă܂Ƃ
{
	const int outer = rule.outerInvalidLines;
	const Board * board = game.GetBoard();
	VirtualStraightData vstrs( board, rule.lineLength + 1/*ԒAp*/ );
	for ( std::set<VirtualStraightData::VirtualStraight>::const_iterator vstr = vstrs.vstraights.begin(); vstr != vstrs.vstraights.end(); vstr++ ) {
		const Board::EStone me = vstr->color;
		const Board::EStone other = ~me;
		//  stop ܂ĂāAVԂW
		if ( vstr->firststop & other && vstr->laststop & other ) {
			if ( vstr->length < rule.lineLength || (me == Board::EStoneBlack && vstr->length > rule.lineLength ) )
				continue;	// ׂ鉿l͖
		}
		
		// spaces ̂Aʒu𒲂ׂ
		const VirtualStraightData::delta delta = VirtualStraightData::GetDelta( vstr->dir );
		const VirtualStraightData::delta sideDelta = delta.rot();
		int myInfluencePoints = 0;
		std::vector< Position > weekPoints;
		if ( vstr->dir == StraightData::EDirVertical || vstr->dir == StraightData::EDirHolizontal ) {
			// ΂߂Ŗ΁AAɒp̖̐̕΂āA`΋
			for ( std::vector< StraightData::Position >::const_iterator spc = vstr->spaces.begin(); spc != vstr->spaces.end(); spc++ ) {
				int cntMyside = 0;
				int cntHisside = 0;
				Board::EStone s = get( board, *spc + sideDelta );
				if      ( s == me )		cntMyside++;
				else if ( s & other )	cntHisside++;
				s = get( board, *spc - sideDelta );
				if      ( s == me )		cntMyside++;
				else if ( s & other )	cntHisside++;
				if ( cntMyside > cntHisside )
					myInfluencePoints++;
				else
					weekPoints.push_back( *spc );
			}
		}
		else {
			// ΂߂ȂA͂̐ΐK
			for ( std::vector< StraightData::Position >::const_iterator spc = vstr->spaces.begin(); spc != vstr->spaces.end(); spc++ ) {
				int mt = 0, mx = 0, ht = 0, hx = 0;
				Board::EStone s;
				int x = spc->x, y = spc->y;
				s = board->Get(x-1,y-1);	{ mx += (s == me);	hx += (s & other); }
				s = board->Get(x  ,y-1);	{ mt += (s == me);	ht += (s & other); }
				s = board->Get(x+1,y-1);	{ mx += (s == me);	hx += (s & other); }
				s = board->Get(x-1,y  );	{ mt += (s == me);	ht += (s & other); }
				s = board->Get(x+1,y  );	{ mt += (s == me);	ht += (s & other); }
				s = board->Get(x-1,y+1);	{ mx += (s == me);	hx += (s & other); }
				s = board->Get(x  ,y+1);	{ mt += (s == me);	ht += (s & other); }
				s = board->Get(x+1,y+1);	{ mx += (s == me);	hx += (s & other); }
				if ( (mt*2+mx) - (ht*2+hx) > 2 )
					myInfluencePoints++;
				else
					weekPoints.push_back( *spc );
			}
		}
		if ( vstr->color == game.OrderToStone(myside) ) {
			// LȎgH
		}
		else {
			// ~߂ɂ
			// @@@ 0 ł͂ȂAintmin ŏׂ
			int wps = 0;		// weekpoint ̕]l
			int sps = 0;		// _̕]l
			int sts = 0;		// stop ̕]l
			if      ( vstr->length < 5 ) {
				wps = 2;
				sts = 4;
			}
			else if ( vstr->length == 5 ) {	// 󂢂ĂΕЕ~߂
				if ( vstr->firststop & other || vstr->laststop & other ) {
					if ( weekPoints.size() >= 2 ) {		// アT
						wps = 4;
						sts = 4;
					}
					else {								// キȂT
						wps = 12;
						sts = 16;
					}
				}
				else {
					if ( weekPoints.size() <= 1 ) {
						sts = 64;
					}
					else {
						sts = 16;
						wps = 8;
					}
				}
			}
			else if ( vstr->length == 6 ) {	// 󂢂Ă΁Aア_ɑł
				if ( ! (vstr->firststop & other || vstr->laststop & other) ) {
					// 󂢂Ăi_ 8, ア_ 24j
					wps = 48;
					sps = 8;
				}
				else {									// Б̋
					if ( weekPoints.size() <= 2 ) {			// ~߂Ȃƍ
						sts = 48;
					}
					else {									// P\邩
						sts = 32;
						wps = 16;
					}
				}
			}
			else {							// _ɑł
				wps = 100;
				sps = 64;
			}
			{
				StraightData::Position pos = vstr->firstend;
				for ( int lpc = 0; vstr->length > lpc; lpc++ ) {
					if ( ! pos.isOuterOf( outer ) )
						scores[ pos.y ][ pos.x ] = std::max( sps, scores[ pos.y ][ pos.x ] );
					pos += delta;
				}
				for ( std::vector< Position >::const_iterator itor = weekPoints.begin(); itor != weekPoints.end(); itor++ ) {
					if ( ! itor->isOuterOf( outer ) )
						scores[ itor->y ][ itor->x ] = std::max( wps, scores[ itor->y ][ itor->x ] );
				}
				Position fp = vstr->firstStopPosition();
				Position lp = vstr->lastStopPosition();
				if ( ! fp.isOuterOf( outer ) )
					scores[ fp.y ][ fp.x ] = std::max( sts, scores[ fp.y ][ fp.x ] );
				if ( ! lp.isOuterOf( outer ) )
					scores[ lp.y ][ lp.x ] = std::max( sts, scores[ lp.y ][ lp.x ] );
			}
		}
	}
}



// @W𓾂
void AI::makeLegalMoves( std::set< Position > & legalMoves, IRENSEI_DEBUG_BOARD  & legalbuf ) const
{
	memset( &legalbuf, -128, sizeof(legalbuf.stones) );
	for ( int y = 0; 19 > y; y++ ) {
		for ( int x = 0; 19 > x; x++ ) {
			legalbuf.stones[y][x] = game.CheckLegalMove( game.GetBoard(), x, y, myside );
			if ( legalbuf.stones[y][x] <= Game::EMResultWin )		// @H
				legalMoves.insert( Position(x,y) );
		}
	}
}

void AI::makeStraightLength( IRENSEI_DEBUG_BOARD & straightLength, const StraightData & straights ) const
{
	memset( straightLength.stones, -128, sizeof(straightLength.stones) );
	for ( std::set<StraightData::Straight>::const_iterator itor = straights.straights.begin(); itor != straights.straights.end(); itor++ ) {
		int x = itor->firstend.x;
		int y = itor->firstend.y;
		StraightData::delta d = StraightData::GetDelta( itor->dir );
		for ( int lpc = 0; itor->length > lpc; lpc++ ) {
			if ( straightLength.stones[ y ][ x ] < itor->length )
				straightLength.stones[ y ][ x ] = itor->length;
			x += d.x;
			y += d.y;
		}
	}
}



void AI::makeVirtualStraightLength( IRENSEI_DEBUG_BOARD & virtualLength, const VirtualStraightData & vstrs, Game::EBlackOrWhite side ) const
{
	memset( virtualLength.stones, -128, sizeof(straightLength.stones) );
	for ( std::set<VirtualStraightData::VirtualStraight>::const_iterator itor = vstrs.vstraights.begin(); itor != vstrs.vstraights.end(); itor++ ) {
		int x = itor->firstend.x;
		int y = itor->firstend.y;
		VirtualStraightData::delta d = VirtualStraightData::GetDelta( itor->dir );
		for ( int lpc = 0; itor->length > lpc; lpc++ ) {
			if ( virtualLength.stones[ y ][ x ] < itor->length )
				virtualLength.stones[ y ][ x ] = itor->length;
			x += d.x;
			y += d.y;
		}
	}
}



int AI::randomMove( IRENSEI_BOARD_POSITION & result, const IRENSEI_DEBUG_BOARD & legalbuf, HANDLE hout ) const
{
	// @̗
	std::vector< Position > legalMoves;
	legalMoves.reserve( 19*19 );
	for ( int y = 0; 19 > y; y++ ) {
		for ( int x = 0; 19 > x; x++ ) {
			// @H
			if ( Game::EMResultWin >= legalbuf.stones[y][x] )
				legalMoves.push_back( Position(x,y) );
		}
	}

	// @肪Γ
	if ( legalMoves.size() == 0 )
		return 0;

	// _őI
	{
		// Ȃׂɋ߂ł悤ɂ
		std::vector< int > cofs;	// d݂̗ݐ
		int total = 0;
		for ( std::vector<Position>::const_iterator itor = legalMoves.begin(); itor != legalMoves.end(); itor++ ) {
			int cof = abs(itor->x - 9) + abs(itor->y - 9);
			cof = 20 - cof;		// 2..20 ŁA傫ȂB܂A[ 0 ɂ͂ȂȂ
			cof = cof * cof;	// Q悵āAW傫
			total += cof;
			cofs.push_back(total);
		}
		int r = total * rand() / (RAND_MAX + 1);
		std::vector< int >::const_iterator i = std::lower_bound( cofs.begin(), cofs.end(), r );
		*reinterpret_cast<Position*>(&result) = legalMoves[ i - cofs.begin() ];
	}

	return 1;
}





//																		//
//						OC^[tF[X							//

// `h
static int ai_undo( int param, IRENSEI_BOARD_POSITION * lastMove, int hout )
{
	return reinterpret_cast<AI*>( param )->Undo( *lastMove, reinterpret_cast<HANDLE>(hout) );
}

static int ai_debug( int param, IRENSEI_DEBUG_BOARD * dest, int idx, int hout )
{
	return reinterpret_cast<AI*>( param )->Debug( *dest, idx, reinterpret_cast<HANDLE>(hout) );
}



// ア`h

// 					`h					//
static int strong_ai_initialize(const IRENSEI_RULE_DESCRIPTION * rule, int side, int hout)
{
	if ( ! CheckIrenseiRuleDescriptionValidity( rule ) )
		return 0;

	// @@@ AI ̃xݒ
	return reinterpret_cast<int>( new AI( *rule, static_cast<Game::EBlackOrWhite>(side) ) );
}
static int strong_ai_finalize(int param, int hout)
{
	delete reinterpret_cast<AI*>( param );
	return 1;
}
static int strong_ai_think(int param, IRENSEI_BOARD_POSITION * result, const IRENSEI_BOARD_POSITION * lastMove, int hout)
{
	return reinterpret_cast<AI*>( param )->Think( *result, lastMove, reinterpret_cast<HANDLE>(hout) );
}






extern void MakeStrongAiDesc( IRENSEI_AI_DESC * desc )
{
	desc->firstVersion = 100;
	desc->lastVersion = 100;
	desc->countDebug = AI::CountDebug();
	strncpy( desc->name, "Rs[^", IRENSEI_AI_NAME_LEN - 1 );
	desc->name[ IRENSEI_AI_NAME_LEN - 1 ] = 0;
	desc->initialize = strong_ai_initialize;
	desc->finalize = strong_ai_finalize;
	desc->think = strong_ai_think;
	desc->undo = ai_undo;
	desc->debug = ai_debug;
}



}	// namespace Irensei


